# Sheepdog
# Copyright 2013, 2014 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Sheepdog is a utility to run arbitary code on a GridEngine cluster and collect
the results, typically by mapping a set of arguments to one function.

Documentation: http://sheepdog.readthedocs.org

Source code: https://github.com/adamgreig/sheepdog

PyPI: https://pypi.python.org/pypi/Sheepdog

Sheepdog is released under the MIT license, see the LICENSE file for details.
"""

from __future__ import print_function

__author__ = "Adam Greig <adam@adamgreig.com>"
__version__ = "0.2.3"
__version_info__ = tuple([int(d) for d in __version__.split(".")])
__license__ = "MIT License"

import os
import sys
import time
import copy
import string
import socket
import random
import getpass

from sheepdog.server import get_server
from sheepdog.storage import Storage
from sheepdog.deployment import Deployer
from sheepdog.job_file import job_file

from sheepdog import serialisation

default_config = {
    "ssh_port": 22,
    "ssh_user": getpass.getuser(),
    "ssh_keyfile": None,
    "ssh_dir": ".sheepdog",
    "dbfile": "./sheepdog.sqlite",
    "port": None,
    "ge_opts": ["-wd $HOME/.sheepdog/", "-o $HOME/.sheepdog/",
                "-e $HOME/.sheepdog/"],
    "shell": "/usr/bin/python",
    "localhost": socket.getfqdn()
}


session_password = None


def map_async(f, args, config, ns=None):
    """Submit *f* with each of *args* on GridEngine, returning the
       (sheepdog-local) request ID.
       
       For details on *config*, see the documentation at:
       http://sheepdog.readthedocs.org/en/latest/configuration.html
       Or in docs/configuration.rst.

       Optionally *ns* is a dict containing a namespace to execute the function
       in, which may itself contain additional functions.
    """
    global session_password
    if not ns:
        ns = {}

    conf = copy.copy(default_config)
    conf.update(config)

    func_bin = serialisation.serialise_function(f)
    args_bin = serialisation.serialise_args(args)
    namespace_bin = serialisation.serialise_namespace(ns)

    storage = Storage(dbfile=conf['dbfile'])
    storage.initdb()
    request_id = storage.new_request(func_bin, namespace_bin, args_bin)

    if not session_password:
        session_password = ''.join(random.choice(string.ascii_letters)
                                   for _ in range(30))

    server = get_server(conf['port'], session_password, conf['dbfile'])
    port = server.port
    url = "http://{0}:{1}/".format(conf['localhost'], port)

    n_args = len(args)
    jf = job_file(url, session_password, request_id, n_args,
                  conf['shell'], conf['ge_opts'])

    deployer = Deployer(
        conf['host'], conf['ssh_port'], conf['ssh_user'], conf['ssh_keyfile'])
    deployer.deploy(jf, request_id, conf['ssh_dir'])
    deployer.submit(request_id, conf['ssh_dir'])

    return request_id

def get_results(request_id, dbfile, block=True, verbose=False):
    """Fetch results for *request_id*. If *block* is true, wait until all the
    results are in. Otherwise, return just what has been received so far.

    If *verbose* is true, print a status message every second with the current
    number of results.

    Returns a list of (arg, result) tuples.

    Where an error occured or no result has been submitted yet, result will be
    None.
    """
    storage = Storage(dbfile=dbfile)
    n_args = storage.count_tasks(request_id)
    n_results = 0
    last_count = None
    while True:
        n_results = storage.count_results(request_id)
        n_errors = storage.count_errors(request_id)
        if verbose and n_results + n_errors != last_count:
            print("{}/{} results, {} errors\r".format(
                  n_results, n_args, n_errors))
            sys.stdout.flush()
        last_count = n_results + n_errors
        if not block or n_results + n_errors == n_args:
            break
        time.sleep(1)

    results = []
    for r in storage.get_tasks_with_results(request_id):
        r1 = r[1] if r[1] is None else serialisation.deserialise_pickle(r[1])
        results.append((serialisation.deserialise_pickle(r[0]), r1))

    return results

def get_errors(request_id, dbfile):
    """Fetch all the errors returned so-far for *request_id*."""
    storage = Storage(dbfile=dbfile)
    errors = []
    for error in storage.get_errors(request_id):
        errors.append((serialisation.deserialise_pickle(error[0]), error[1]))
    return errors

def map(f, args, config, ns=None, verbose=True):
    """Submit *f* with each of *args* on GridEngine, wait until all the results
       are in, and return them in the same order as *args*. If an error occured
       for an arg, None is returned in that position. Call `get_errors` to get
       details on the errors that occured.
       
       For details on *config*, see the documentation at:
       http://sheepdog.readthedocs.org/en/latest/configuration.html
       Or in docs/configuration.rst.

       Optionally *ns* is a dict containing a namespace to execute the function
       in, which may itself contain additional functions.

       If *verbose* is true, print out how many results are in so-far while
       waiting.
    """
    request_id = map_async(f, args, config, ns)
    if verbose:
        print("Deployed with request ID", request_id)

    conf = copy.copy(default_config)
    conf.update(config)
    results = get_results(request_id, conf['dbfile'], block=True, verbose=True)
    storage = Storage(dbfile=conf['dbfile'])

    if verbose and storage.count_errors(request_id) != 0:
        print("Some errors occured, view them with get_errors({}, '{}')"
              .format(request_id, conf['dbfile']), file=sys.stderr)

    return [r[1] for r in results]



# Sheepdog
# Copyright 2013 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Sheepdog's HTTP server endpoints.

The Server class sets up a server on another subprocess, ready to receive
requests from workers. Uses Tornado if available, else falls back to the Flask
debug web server.
"""

import json
import atexit
import socket
from functools import wraps
from multiprocessing import Process
from flask import Flask, Response, request, g
from sheepdog.storage import Storage

try:
    from tornado.wsgi import WSGIContainer
    from tornado.httpserver import HTTPServer
    from tornado.ioloop import IOLoop
    USE_TORNADO = True
except ImportError:
    USE_TORNADO = False

app = Flask(__name__)

def check_auth(username, password):
    return username == 'sheepdog' and password == app.config['PASSWORD'] 

def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.authorization
        if not auth or not check_auth(auth.username, auth.password):
            return Response("Unauthorized", 401)
        return f(*args, **kwargs)
    return decorated

@app.route('/', methods=['GET'])
@requires_auth
def get_config():
    """Endpoint for workers to fetch their configuration before execution.
       Workers should specify `request_id` (integer) and `job_index` (integer)
       from their job file.

       Returns a JSON object:
       
       {"func": (serialised function object),
        "args": (serialised arguments list)
       }

       with HTTP status 200 on success.
    """
    storage = get_storage()
    request_id = int(request.args['request_id'])
    job_index = int(request.args['job_index'])
    details = storage.get_details(request_id, job_index)
    func = details[0].decode()
    ns = details[1].decode()
    args = details[2].decode()
    return json.dumps({"func": func, "ns": ns, "args": args})

@app.route('/', methods=['POST'])
@requires_auth
def submit_result():
    """Endpoint for workers to submit results arising from successful function
       execution. Should specify `request_id` (integer), `job_index` (integer)
       and `result` (serialised result) HTTP POST parameters.

       Returns the string "OK" and HTTP 200 on success.
    """
    storage = get_storage()
    request_id = int(request.form['request_id'])
    job_index = int(request.form['job_index'])
    result = request.form['result'].encode()
    storage.store_result(request_id, job_index, result)
    return "OK"

@app.route('/error', methods=['POST'])
@requires_auth
def report_error():
    """Endpoint for workers to report back errors in function execution.
       Workers should specify `request_id` (integer), `job_index` (integer) and
       `error` (an error string) HTTP POST parameters.

       Returns the string "OK" and HTTP 200 on success.
    """
    storage = get_storage()
    request_id = int(request.form['request_id'])
    job_index = int(request.form['job_index'])
    error = str(request.form['error'])
    storage.store_error(request_id, job_index, error)
    return "OK"

def get_storage():
    """Retrieve the request-local database connection, creating it if required.
    """
    if not hasattr(g, '_storage'):
        dbfile = app.config['DBFILE']
        g._storage = Storage(dbfile)
    return g._storage

def _get_free_port():
    """Get a port that should be free on the system."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    return port

def run_server(port, password, dbfile):
    """Start up the HTTP server. If Tornado is available it will be used, else
       fall back to the Flask debug server.
    """
    app.config['PASSWORD'] = password
    app.config['DBFILE'] = dbfile

    if USE_TORNADO:
        # When running inside an IPython Notebook, the IOLoop
        # can inherit a stale instance from the parent process,
        # so clear that.
        # https://github.com/adamgreig/sheepdog/issues/15
        # Thanks @minrk!
        if hasattr(IOLoop, '_instance'):
            del IOLoop._instance
        IOLoop.clear_current()

        HTTPServer(WSGIContainer(app)).listen(port)
        IOLoop.instance().start()
    else:
        app.run(host='0.0.0.0', port=port)


class Server:
    """Run the HTTP server for workers to request arguments and return results.

       Should be used via get_server(port, password, dbfile) to manage servers
       running globally.
    """

    def __init__(self, port, password, dbfile):
        """__init__ creates and starts the HTTP server.
        """
        if not port:
            port = _get_free_port()
        self.port = port
        self.password = password
        self.dbfile = dbfile
        self.server = Process(target=run_server, args=(port, password, dbfile))
        self.server.start()

    def stop(self):
        """Terminate the HTTP server."""
        self.server.terminate()
        self.server.join()

    def __del__(self):
        self.stop()

_servers = {}
def get_server(port, password, dbfile):
    """Either start a new server or retrieve a reference to an existing server.
       Only one server may run per port. If the server currently running on
       that port has a different password or dbfile, a RuntimeError is raised.

       If `None` is specified for port, a port is picked randomly and that
       server is the one referenced for `None` thereafter.
    """
    global _servers
    if port not in _servers:
        server = Server(port, password, dbfile)
        _servers[port] = server

        # When port is None, this adds another entry for the server on the port
        # it's actually running on
        _servers[server.port] = server
    else:
        server = _servers[port]
        if server.password != password or server.dbfile != dbfile:
            raise RuntimeError("A server is already running on that port with "
                               "a different password or dbfile.")
    return server

def _cleanup_servers():
    """Shut down all running servers in _servers"""
    global _servers
    for port in _servers:
        _servers[port].stop()
    del _servers

# The servers will continue in their subprocess without going out of scope
# if we hold a reference in _servers, so register a handler to get rid of
# them when this process ends.
atexit.register(_cleanup_servers)



# Sheepdog
# Copyright 2013, 2014 Adam Greig
#
# Released under the MIT license. See LICENSE file for details.

"""
Code for deploying code to servers and executing jobs on GridEngine.
"""

import os.path
import paramiko

class Deployer:
    """Connect to a remote SSH server, copy a file over, run qsub."""

    def __init__(self, host, port, user, keyfile=None):
        """__init__ takes (host, port, user, keyfile) to specify which SSH
           server to connect to and how to connect to it.
        """
        self.host = host
        self.port = port
        self.user = user
        self.keyfile = keyfile
        self.sock = None

        self.ssh = paramiko.SSHClient()
        self.ssh.load_system_host_keys()
        self._process_config()

        self.ssh.connect(self.host, self.port, self.user,
                         key_filename=keyfile, sock=self.sock)

    def deploy(self, jobfile, request_id, directory):
        """Copy *jobfile* (a string of the file contents) to the connected
           remote host, placing it in *directory* with a filename containing
           *request_id*.
        """
        path = self._get_jobfile_path(request_id, directory)
        sftp = self.ssh.open_sftp()

        try:
            sftp.mkdir(directory, mode=0o750)
        except (IOError, OSError):
            pass

        with sftp.open(path, 'w') as f:
            f.write(jobfile)

    def submit(self, request_id, directory):
        """Submit a job to the GridEngine cluster on the connected remote host.
           Calls qsub with the job identified by request_id and directory.
        """
        path = self._get_jobfile_path(request_id, directory)
        si, so, se = self.ssh.exec_command("qsub {}".format(path))
        return so.read()

    def _get_jobfile_path(self, request_id, directory):
        """Put together the path at which the job file for a given *request_id*
           will be found, given *directory*.
        """
        return os.path.join(directory, "sheepdog_{:03}.py".format(request_id))

    def _process_config(self):
        """Parse the user's local ssh config (if any) and use it to update
           the hostname, port and username used for connection.
        """
        config = paramiko.SSHConfig()
        config_path = os.path.expanduser("~/.ssh/config")
        
        try:
            with open(config_path) as f:
                config.parse(f)
        except IOError:
            pass

        host_config = config.lookup(self.host)

        self.host = host_config.get('hostname', self.host)
        self.user = host_config.get('user', self.user)
        self.port = host_config.get('port', self.port)

        if 'proxycommand' in host_config:
            self.sock = paramiko.ProxyCommand(host_config['proxycommand'])
